#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#define _BSD_SOURCE 1
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

static char const prog[]="statfile";

#define set(var, val) \
  ((void)(pathexec_env(1+(var), (val))==0 && \
          (die6sys(111, prog, fatal, err_unable, err_setvar, (var), err_), 0)))

int main(int argc, char** argv) {
  struct stat statbuf;
  char buf[(sizeof (unsigned long)*CHAR_BIT+2)/3+1];
  int do_dev=0;
  int do_ino=0;
  int do_type=0;
  int do_mode=0;
  int do_nlink=0;
  int do_uid=0;
  int do_gid=0;
  int do_rdev=0;
  int do_size=0;
  int do_blksize=0;
  int do_blocks=0;
  int do_atime=0;
  int do_mtime=0;
  int do_ctime=0;

  if (argc<4)
    die4x(111, prog, usage, prog, " file options command [arg ...]");

  {
    char const* x=argv[1];
    while (*x!='\0') {
      switch (*x) {
        case 'd': do_dev=1;     break;
        case 'i': do_ino=1;     break;
        case 't': do_type=1;    break;
        case 'm': do_mode=1;    break;
        case 'n': do_nlink=1;   break;
        case 'u': do_uid=1;     break;
        case 'g': do_gid=1;     break;
        case 'r': do_rdev=1;    break;
        case 's': do_size=1;    break;
        case 'S': do_blksize=1; break;
        case 'b': do_blocks=1;  break;
        case 'A': do_atime=1;   break;
        case 'M': do_mtime=1;   break;
        case 'C': do_ctime=1;   break;
        default: {
          char opt[2]={ 0, 0 };
          *opt=*x;
          die5x(111, prog, fatal, "unrecognized option", err_, opt);
        }
      }
      ++x;
    }
  }

  if (lstat(argv[2], &statbuf)!=0)
    die6sys(111, prog, fatal, err_unable, err_stat, argv[2], err_);

  #define doit(member, var) do { \
    if (statbuf.st_##member<0 || \
        statbuf.st_##member!=(unsigned long)statbuf.st_##member) \
      die5x(100, prog, fatal, #member, err_oflow, "unsigned long"); \
    if (do_##member) { \
      buf[fmt_ulong(buf, statbuf.st_##member)]='\0'; \
      set(var, buf); \
    } \
  } while (0)

  doit(dev, "$DEV");
  doit(ino, "$INO");
  doit(nlink, "$NLINK");
  doit(uid, "$UID");
  doit(gid, "$GID");
  doit(rdev, "$RDEV");
  doit(size, "$SIZE");
  doit(blksize, "$BLKSIZE");
  doit(blocks, "$BLOCKS");
  doit(atime, "$ATIME");
  doit(mtime, "$MTIME");
  doit(ctime, "$CTIME");

  if (do_type) {
    char const* type;
    if (S_ISLNK(statbuf.st_mode)) type="l"; else
    if (S_ISREG(statbuf.st_mode)) type="-"; else
    if (S_ISDIR(statbuf.st_mode)) type="d"; else
    if (S_ISCHR(statbuf.st_mode)) type="c"; else
    if (S_ISBLK(statbuf.st_mode)) type="b"; else
    if (S_ISFIFO(statbuf.st_mode)) type="p"; else
    if (S_ISSOCK(statbuf.st_mode)) type="s"; else
    type="unknown";
    set("$TYPE", type);
  }

  if (do_mode) {
    unsigned long const mode=statbuf.st_mode&07777;
    buf[0]='0';
    buf[1]='0'+((mode>>9)&07);
    buf[2]='0'+((mode>>6)&07);
    buf[3]='0'+((mode>>3)&07);
    buf[4]='0'+(mode&07);
    buf[5]='\0';
    set("$MODE", buf);
    set("$MODE_SUID",   (statbuf.st_mode&04000? "1": 0));
    set("$MODE_SGID",   (statbuf.st_mode&02000? "1": 0));
    set("$MODE_STICKY", (statbuf.st_mode&01000? "1": 0));
    set("$MODE_RUSR",   (statbuf.st_mode&00400? "1": 0));
    set("$MODE_WUSR",   (statbuf.st_mode&00200? "1": 0));
    set("$MODE_XUSR",   (statbuf.st_mode&00100? "1": 0));
    set("$MODE_RGRP",   (statbuf.st_mode&00040? "1": 0));
    set("$MODE_WGRP",   (statbuf.st_mode&00020? "1": 0));
    set("$MODE_XGRP",   (statbuf.st_mode&00010? "1": 0));
    set("$MODE_ROTH",   (statbuf.st_mode&00004? "1": 0));
    set("$MODE_WOTH",   (statbuf.st_mode&00002? "1": 0));
    set("$MODE_XOTH",   (statbuf.st_mode&00001? "1": 0));
  }

  argv+=3;
  pathexec(argv);
  die6sys(111, prog, fatal, err_unable, err_exec, argv[0], err_);
  return 111;
}
